@@ -199,6 +199,7 @@ const char *ata1_cfg_str[] = {			/* word 0 in ATA-1 mode */
 /* words 47 & 59: sector_xfer_max & sector_xfer_cur */
 #define SECTOR_XFER		0x00ff  /* sectors xfered on r/w multiple cmds*/
 #define MULTIPLE_SETTING_VALID  0x0100  /* 1=multiple sector setting is valid */
+#define SANITIZE_FEAT_SUP       0x1000 /* SANITIZE FETURES set is supported */
 
 /* word 49: capabilities 0 */
 #define STD_STBY  		0x2000  /* 1=standard values supported (ATA);
@@ -223,6 +224,26 @@ const char *ata1_cfg_str[] = {			/* word 0 in ATA-1 mode */
 #define OK_W64_70		0x0002  /* see above for word descriptions */
 #define OK_W54_58		0x0001  /* current cyl, head, sector, cap. info valid */
 
+/* word 59: sanitize feature set */
+static const char *feat_word59_str[16] = {
+	"BLOCK_ERASE_EXT command",						/* word 84 bit 15 */
+	"OVERWRITE_EXT command",						/* word 84 bit 14 */
+	"CRYPTO_SCRAMBLE_EXT command",					/* word 59 bit 13 */
+	"SANITIZE feature set",							/* word 59 bit 12 */
+	NULL, 											/* word 59 bit 11 */
+	"SANITIZE_ANTIFREEZE_LOCK_EXT command",			/* word 59 bit 10 */
+	NULL,				/* word 59 bit  9 */
+	NULL,				/* word 59 bit  8 */
+	NULL,				/* word 59 bit  7 */
+	NULL,				/* word 59 bit  6 */
+	NULL,				/* word 59 bit  5 */
+	NULL,				/* word 59 bit  4 */
+	NULL,				/* word 59 bit  3 */
+	NULL,				/* word 59 bit  2 */
+	NULL,				/* word 59 bit  1 */
+	NULL				/* word 59 bit  0 */
+};
+
 /*word 63,88: dma_mode, ultra_dma_mode*/
 #define MODE_MAX		7	/* bit definitions force udma <=7 (when
 					 * udma >=8 comes out it'll have to be
@@ -325,10 +346,6 @@ const char actual_ver[MINOR_MAX+2] = {
 #define SUPPORT_48_BIT		0x0400  
 #define NUM_CMD_FEAT_STR	48
 
-static const char unknown[8] = "obsolete";
-//static const char unknown[8] = "unknown";
-#define unknown "unknown-"
-
 static const char *feat_word69_str[16] = { 
 	"CFast specification support",			/* word 69 bit 15 */
 	"Deterministic read data after TRIM",		/* word 69 bit 14 */
@@ -338,12 +355,12 @@ static const char *feat_word69_str[16] = {
 	"WRITE BUFFER DMA command",			/* word 69 bit 10 */
 	"SET MAX SETPASSWORD/UNLOCK DMA commands",	/* word 69 bit  9 */
 	"DOWNLOAD MICROCODE DMA command",		/* word 69 bit  8 */
-	"reserved 69[7]",				/* word 69 bit  7 */
-	"reserved 69[6]",				/* word 69 bit  6 */
-	"Deterministic read ZEROs after TRIM",		/* word 69 bit  5 */
-	"reserved 69[4]",				/* word 69 bit  4 */
-	"reserved 69[3]",				/* word 69 bit  3 */
-	"reserved 69[2]",				/* word 69 bit  2 */
+	"IEEE 1667 authentication of transient storage devices",	/* word 69 bit  7 */
+	"Optional ATA device 28-bit commands",			/* word 69 bit  6 */
+	"Deterministic read ZEROs after TRIM",			/* word 69 bit  5 */
+	"Device encrypts all user data",				/* word 69 bit  4 */
+	"Extended number of user addressable sectors ",	/* word 69 bit  3 */
+	"All write cache is non-volatile",				/* word 69 bit  2 */
 	"reserved 69[1]",				/* word 69 bit  1 */
 	"reserved 69[0]",				/* word 69 bit  0 */
 };
@@ -648,8 +665,69 @@ static int is_cfa_dev (__u16 *id)
 	return id[0] == 0x848a || id[0] == 0x844a || (id[83] & 0xc004) == 0x4004;
 }
 
+static void print_devslp_info (int fd, __u16 *id)
+{
+	/* Print DEVSLP information */
+	if (id[78] & 0x0100) {
+		__u8 buf[512];
+		int deto = 0;
+		int mdat = 0;
+
+		memset(buf, 0, 512);
+		if (fd != -1 && !get_log_page_data(fd, 0x30, 8, buf) && (buf[0x37] & 0x80)) {
+			mdat = buf[0x30] & 0x1f;
+			deto = buf[0x31];
+			printf("Device Sleep:\n");
+			printf("\tDEVSLP Exit Timeout (DETO): %d ms (%s)\n", deto?deto:20, deto?"drive":"default");
+			printf("\tMinimum DEVSLP Assertion Time (MDAT): %d ms (%s)\n", mdat?mdat:10, deto?"drive":"default");
+		}
+	}
+}
+
+static void
+print_logical_sector_sizes (int fd, unsigned int current_lss)
+{
+	__u8 d[512] = {0,};
+	int i, found = 0, rc, found_byte_lss = 0, found_word_lss = 0, lss_is_bytes = 0;
+
+	rc = get_log_page_data(fd, 0x2f, 0, d);
+	if (rc)
+		return;
+	/*
+	 * Some devices incorrectly return logical sector size (lss) as bytes.
+	 * Scan log entries and if the current lss is in the log, and there isn't an equivalent
+	 * entry in words, assume the device is returning the logical sector size in bytes.
+	 * Doing this takes two passes through the log data.
+	 */
+	for (i = 0; i < 128; i += 16) {
+		unsigned int lss;
+		if ((d[i] & 0x80) == 0)  /* Is this descriptor valid? */
+			continue;  /* not valid */
+		lss = d[i + 4] | (d[i + 5] << 8) | (d[i + 6] << 16) | (d[i + 7] << 24);  /* logical sector size */
+		if (lss == current_lss) {
+			found_byte_lss = 1;  /* found lss in bytes */
+		} else if ((lss * 2) == current_lss) {
+			found_word_lss = 1;  /* found lss in words */
+			break;
+		}
+	}
+	lss_is_bytes = !found_word_lss && found_byte_lss;
+	for (i = 0; i < 128; i += 16) {
+		unsigned int lss;
+		if ((d[i] & 0x80) == 0)  /* Is this descriptor valid? */
+			continue;  /* not valid */
+		if (!found++)
+			printf(" [ Supported:");
+		lss = d[i + 4] | (d[i + 5] << 8) | (d[i + 6] << 16) | (d[i + 7] << 24);  /* logical sector size */
+		lss = lss_is_bytes ? lss : (lss * 2);
+		printf(" %u", lss);
+	}
+	if (found)
+		printf(" ]");
+}
+
 /* our main() routine: */
-void identify (__u16 *id_supplied)
+void identify (int fd, __u16 *id_supplied)
 {
 	unsigned int sector_bytes = 512;
 	__u16 val[256], ii, jj, kk;
@@ -910,19 +988,19 @@ void identify (__u16 *id_supplied)
 					if(llabs((long long)(mm - bb)) > llabs((long long)(nn - bb)))
 						mm = nn;
 				}
-				printf("\tCHS current addressable sectors:%11u\n",mm);
+				printf("\tCHS current addressable sectors:%12u\n",mm);
 			} 
 		}
 		if (val[CAPAB_0] & LBA_SUP) {
 		/* LBA addressing */
-			printf("\tLBA    user addressable sectors:%11u\n",ll);
+			printf("\tLBA    user addressable sectors:%12u\n",ll);
 			if( ((val[CMDS_SUPP_1] & VALID) == VALID_VAL) &&
 			     (val[CMDS_SUPP_1] & SUPPORT_48_BIT) ) {
 				bbbig = (__u64)val[LBA_64_MSB] << 48 | 
 				        (__u64)val[LBA_48_MSB] << 32 |
 				        (__u64)val[LBA_MID] << 16 | 
 					val[LBA_LSB] ;
-				printf("\tLBA48  user addressable sectors:%11llu\n", (unsigned long long)bbbig);
+				printf("\tLBA48  user addressable sectors:%12llu\n", (unsigned long long)bbbig);
 			}
 		}
 		if((val[106] & 0xc000) != 0x4000) {
@@ -934,20 +1012,23 @@ void identify (__u16 *id_supplied)
 			if (val[106] & (1<<12))
 				lsize = (val[118] << 16) | val[117];
 			sector_bytes = 2 * lsize;
-			printf("\t%-31s %11u bytes\n","Logical  Sector size:", sector_bytes);
+			printf("\t%-31s %11u bytes","Logical  Sector size:", sector_bytes);
+			print_logical_sector_sizes(fd, sector_bytes);
+			printf("\n");
 			printf("\t%-31s %11u bytes\n","Physical Sector size:", sector_bytes * pfactor);
 			if ((val[209] & 0xc000) == 0x4000) {
 				unsigned int offset = val[209] & 0x1fff;
 				printf("\t%-31s %11u bytes\n", "Logical Sector-0 offset:", offset * sector_bytes);
 			}
 		}
-		if (!bbbig) bbbig = (__u64)(ll>mm ? ll : mm); /* # 512 byte blocks */
+		if (!bbbig) bbbig = (__u64)((ll > mm) ? ll : mm); /* # 512 byte blocks */
 		if (!bbbig) bbbig = bb;
-		bbbig *= (sector_bytes / 512);
-		printf("\tdevice size with M = 1024*1024: %11llu MBytes\n", (unsigned long long)(bbbig>>11));
-		bbbig = (bbbig<<9)/1000000;
-		printf("\tdevice size with M = 1000*1000: %11llu MBytes ", (unsigned long long)bbbig);
-		if(bbbig > 1000) printf("(%llu GB)\n", (unsigned long long)(bbbig/1000));
+		bbbig *= sector_bytes;
+
+		printf("\tdevice size with M = 1024*1024: %11llu MBytes\n", bbbig / (1024ull * 1024ull));
+		bbbig /= 1000ull;
+		printf("\tdevice size with M = 1000*1000: %11llu MBytes ",  bbbig / 1000ull);
+		if (bbbig > 1000ull) printf("(%llu GB)\n", bbbig/1000000ull);
 		else printf("\n");
 	}
 
@@ -1155,8 +1236,10 @@ void identify (__u16 *id_supplied)
 
 	/* Programmed IO stuff */
 	printf("\tPIO: ");
-        /* If a drive supports mode n (e.g. 3), it also supports all modes less
-	 * than n (e.g. 3, 2, 1 and 0).  Print all the modes. */
+	/*
+	 * If a drive supports mode n (e.g. 3), it also supports all modes less
+	 * than n (e.g. 3, 2, 1 and 0).  Print all the modes.
+	 */
 	if((val[WHATS_VALID] & OK_W64_70) && (val[ADV_PIO_MODES] & PIO_SUP)) {
 		jj = ((val[ADV_PIO_MODES] & PIO_SUP) << 3) | 0x0007;
 		for (ii = 0; ii <= PIO_MODE_MAX ; ii++) {
@@ -1204,6 +1287,8 @@ void identify (__u16 *id_supplied)
 			print_features(val[SATA_SUPP_0], val[SATA_EN_0], feat_sata0_str);
 		if (val[SCT_SUPP] & 0x1)
 			print_features(val[SCT_SUPP], val[SCT_SUPP] & 0x3f, feat_sct_str);
+		if (val[SECTOR_XFER_CUR] & SANITIZE_FEAT_SUP)
+			print_features(val[SECTOR_XFER_CUR], val[SECTOR_XFER_CUR], feat_word59_str);
 	}
 	if (like_std > 6) {
 		const __u16 trimd = 1<<14;	/* deterministic read data after TRIM */
@@ -1303,50 +1388,62 @@ void identify (__u16 *id_supplied)
 		}
 	}
 
-	if((val[RM_STAT] & RM_STAT_BITS) == RM_STAT_SUP) 
+	if((val[RM_STAT] & RM_STAT_BITS) == RM_STAT_SUP)
 		printf("\t\tRemovable Media Status Notification feature set supported\n");
 
+
 	/* security */
-	if((val[CMDS_SUPP_2] & 2) && (eqpt != CDROM) && (like_std > 3) && (val[SECU_STATUS] || val[ERASE_TIME] || val[ENH_ERASE_TIME]))
+	if ((val[CMDS_SUPP_0] & (1 << 1)) && (eqpt != CDROM) && (like_std > 3) && (val[SECU_STATUS] || val[ERASE_TIME] || val[ENH_ERASE_TIME]))
 	{
 		printf("Security: \n");
-		if(val[PSWD_CODE] && (val[PSWD_CODE] != 0xffff))
-			printf("\tMaster password revision code = %u\n",val[PSWD_CODE]);
+		if (val[PSWD_CODE] && (val[PSWD_CODE] != 0xffff))
+			printf("\tMaster password revision code = %u\n", val[PSWD_CODE]);
 		jj = val[SECU_STATUS];
-		if(jj) {
+		if (jj) {
 			for (ii = 0; ii < NUM_SECU_STR; ii++) {
-				if(!(jj & 0x0001)) printf("%s", ii ? "\tnot\t" : "\t(?)\t");
-				else		   printf("\t\t");
-				printf("%s\n",secu_str[ii]);
-				jj >>=1;
+				if (!(jj & 0x0001)) printf("%s", ii ? "\tnot\t" : "\t(?)\t");
+				else                printf("\t\t");
+				printf("%s\n", secu_str[ii]);
+				jj >>= 1;
 			}
-			if(val[SECU_STATUS] & SECU_ENABLED) {
+			if (val[SECU_STATUS] & SECU_ENABLED) {
 				printf("\tSecurity level ");
-				if(val[SECU_STATUS] & SECU_LEVEL) printf("maximum\n");
-				else				  printf("high\n");
+				if (val[SECU_STATUS] & SECU_LEVEL) printf("maximum\n");
+				else                               printf("high\n");
 			}
 		}
-		jj =  val[ERASE_TIME];
-		kk =  val[ENH_ERASE_TIME];
-		if((jj && jj <= 0x00ff) || (kk && kk <= 0x00ff)) {
-			printf("\t");
-			if (jj) {
-				if (jj == 0xff)
-					printf("more than 508");
-				else
-					printf("%u", jj * 2);
-				printf("min for SECURITY ERASE UNIT. ");
+		jj = val[ERASE_TIME];                                 // Grab normal erase time
+		unsigned int const ext_time_n = (jj & (1 << 15)) != 0;// Check if erase time is extended format or not (ACS-3)
+		jj = ext_time_n ? jj & 0x7FFF: jj & 0x00FF;           // Mask off reserved bits accordingly
+		kk = val[ENH_ERASE_TIME];                             // Grab enhanced erase time
+		unsigned int const ext_time_e = (kk & (1 << 15)) != 0;// Check if erase time is extended format or not (ACS-3)
+		kk = ext_time_e ? kk & 0x7FFF: kk & 0x00FF;           // Mask off reserved bits accordingly
+		if ((jj != 0) || (kk != 0)) printf("\t");
+		if (jj != 0) {
+			if (ext_time_n && (jj == 0x7FFF)) {
+				printf("more than 65532");
+			} else if (!ext_time_n && (jj == 0x00FF)) {
+				printf("more than 508");
+			} else {
+				printf("%u", jj * 2);
 			}
-			if (kk) {
-				if (kk == 0xff)
-					printf("more than 508");
-				else
-					printf("%u", kk * 2);
-				printf("min for ENHANCED SECURITY ERASE UNIT. ");
-			}
-			printf("\n");
+			printf("min for SECURITY ERASE UNIT.");
 		}
+		if ((jj != 0) && (kk != 0)) printf(" ");
+		if (kk != 0) {
+			if (ext_time_e && (kk == 0x7FFF)) {
+				printf("more than 65532");
+			} else if (!ext_time_e && (kk == 0x00FF)) {
+				printf("more than 508");
+			} else {
+				printf("%u", kk * 2);
+			}
+			printf("min for ENHANCED SECURITY ERASE UNIT.");
+		}
+		printf("\n");
 	}
+
+
 	//printf("w84=0x%04x w87=0x%04x like_std=%d\n", val[84], val[87], like_std);
 	if((eqpt != CDROM) && (like_std > 3) && (val[CMDS_SUPP_2] & WWN_SUP)) {
 		printf("Logical Unit WWN Device Identifier: %04x%04x%04x%04x\n", val[108], val[109], val[110], val[111]);
@@ -1373,6 +1470,7 @@ void identify (__u16 *id_supplied)
 			printf(" determined by CSEL");
 		printf("\n");
 	}
+	print_devslp_info(fd, val);
 
 	/* more stuff from std 5 */
 	if ((like_std > 4) && (eqpt != CDROM)) {
